class NebulaScrapeError(Exception):
    pass

class CloudflareLoopProtection(NebulaScrapeError):
    pass

class CloudflareIUAMError(NebulaScrapeError):
    pass

class CloudflareSolveError(NebulaScrapeError):
    pass

class CloudflareChallengeError(NebulaScrapeError):
    pass

class CloudflareCaptchaError(NebulaScrapeError):
    pass

class CloudflareCaptchaProvider(NebulaScrapeError):
    pass

class CloudflareCode1020(NebulaScrapeError):
    pass

class TransportError(NebulaScrapeError):
    pass

class FingerprintError(NebulaScrapeError):
    pass

class SessionRotationError(NebulaScrapeError):
    pass

class WAFDetectedError(NebulaScrapeError):
    pass

class TurnstileChallengeError(NebulaScrapeError):
    pass

class RetryExhaustedError(NebulaScrapeError):
    pass
